#!/usr/bin/env python3
import numpy as np
import torch
import torchvision.transforms.functional as functional
import torch.nn.functional as F
from models import build_model
from util.misc import nested_tensor_from_tensor_list
import iio
import subprocess
import sys
import os
# Cache model in the currect folder
os.environ['TORCH_HOME'] = os.environ['bin']

class Compose(object):
    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, image):
        for t in self.transforms:
            image = t(image)
        return image

    def __repr__(self):
        format_string = self.__class__.__name__ + "("
        for t in self.transforms:
            format_string += "\n"
            format_string += "    {0}".format(t)
        format_string += "\n)"
        return format_string

class Normalize(object):
    def __init__(self, mean, std):
        self.mean = mean
        self.std = std

    def __call__(self, image):
        image = functional.normalize(image, mean=self.mean, std=self.std)
        return image

class ToTensor(object):
    def __call__(self, img):
        return functional.to_tensor(img)

def resize(image, size, max_size=None):
    # size can be min_size (scalar) or (w, h) tuple
    def get_size_with_aspect_ratio(image_size, size, max_size=None):
        w, h = image_size
        if max_size is not None:
            min_original_size = float(min((w, h)))
            max_original_size = float(max((w, h)))
            if max_original_size / min_original_size * size > max_size:
                size = int(round(max_size * min_original_size / max_original_size))
        if (w <= h and w == size) or (h <= w and h == size):
            return (h, w)
        if w < h:
            ow = size
            oh = int(size * h / w)
        else:
            oh = size
            ow = int(size * w / h)
        return (oh, ow)

    def get_size(image_size, size, max_size=None):
        if isinstance(size, (list, tuple)):
            return size[::-1]
        else:
            return get_size_with_aspect_ratio(image_size, size, max_size)

    size = get_size(image.size, size, max_size)
    rescaled_image = functional.resize(image, size)

    return rescaled_image

class Resize(object):
    def __init__(self, sizes, max_size=None):
        assert isinstance(sizes, (list, tuple))
        self.sizes = sizes
        self.max_size = max_size

    def __call__(self, img):
        size = self.sizes
        return resize(img, size, self.max_size)

img_input = sys.argv[1]
out_name = sys.argv[2]
out_svg = out_name + '.svg'
out_txt = out_name + '.txt'
out_png = out_name + '.png'
out_eps = out_name + '.eps'
path_network = sys.argv[3]
thr_score = float(sys.argv[4])

# obtain checkpoints
checkpoint = torch.load(path_network, map_location='cpu')

# load model
args = checkpoint['args']
setattr(args, 'device', 'cpu')
model, _, postprocessors = build_model(args)
model.load_state_dict(checkpoint['model'])
model.eval()

# load image
raw_img = iio.read(img_input) / 255.
h, w = raw_img.shape[0], raw_img.shape[1]
orig_size = torch.as_tensor([int(h), int(w)])

# normalize image
test_size = 1100
normalize = Compose([
        ToTensor(),
        Normalize([0.538, 0.494, 0.453], [0.257, 0.263, 0.273]),
        #Resize([test_size]),
    ])
img = normalize(raw_img)
inputs = nested_tensor_from_tensor_list([img])

with torch.no_grad():
    outputs = model(inputs)[0]

out_logits, out_line = outputs['pred_logits'], outputs['pred_lines']
prob = F.softmax(out_logits, -1)
scores, labels = prob[..., :-1].max(-1)
img_h, img_w = orig_size.unbind(0)
scale_fct = torch.unsqueeze(torch.stack([img_w, img_h, img_w, img_h], dim=0), dim=0)
lines = out_line * scale_fct[:, None, :]
lines = lines.view(-1, 2, 2)
lines = lines.flip([-1])# this is yxyx format
scores = scores.detach().numpy()
keep = scores >= thr_score
scores = scores[keep]
keep = keep.squeeze()
lines = lines[keep]
if lines.size()[0] > 0:
    lines = lines.reshape(lines.shape[0], -1)


##### OUTPUT ######

ftxt = open(out_txt, "w")
fcsv = open(out_svg, "w")
feps = open(out_eps, "w")

fcsv.write('<?xml version="1.0" standalone="no"?>\n<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN"\n"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd"><svg version="1.1" xmlns="http://www.w3.org/2000/svg"\nxmlns:xlink="http://www.w3.org/1999/xlink"\nwidth="'+str(w)+'" height="'+str(h)+'" viewBox="0 0 '+str(w)+' '+str(h)+'">')

feps.write("%!PS-Adobe-3.0 EPSF-3.0\n");
feps.write(f"%%BoundingBox: 0 0 {w} {h}\n");
feps.write("%%Creator: LSD, Line Segment Detector\n");
feps.write(f"%%Title: ({out_eps})\n");
feps.write("%%EndComments\n");

color='black'
width=1.5
for tp_id, (line, score) in enumerate(zip(lines, scores)):
    y1, x1, y2, x2 = line # this is yxyx
    fcsv.write('<line x1="'+str(x1.item())+'" y1="' +str(y1.item())+ '" x2="' +str(x2.item())+ '" y2="' +str(y2.item())+ '" stroke="'+color+'" stroke-width="'+str(width)+'"/>\n')
    ftxt.write(str(x1.item())+' '+str(y1.item())+' '+str(x2.item())+' '+str(y2.item())+' '+str(score)+'\n')
    feps.write(f"newpath {x1} {h-y1} moveto {x2} {h-y2} lineto 1.5 setlinewidth stroke\n")
fcsv.write('\n</svg>')
feps.write("showpage\n");
feps.write("%%EOF\n");
ftxt.close() 
fcsv.close() 
feps.close()

p2 = ['gs', '-dNOPAUSE', '-dBATCH', '-sDEVICE=pnggray', '-dGraphicsAlphaBits=4', '-r72', \
        '-dEPSCrop', '-sOutputFile='+out_png, out_eps]
val1 = subprocess.run(p2)

im = iio.read(out_png)[:,:,:3]
im = im.mean(axis=-1)
im = 255-im
iio.write(out_png, im)
