import os
import cv2
import torch

from deeplsd.models.deeplsd import DeepLSD

import sys
import iio
import subprocess

import argparse

def main(
    im: str,
    ckpt: str,
    out_txt: str = 'lines.txt',
    out_svg: str = 'lines.svg',
    out_eps: str = 'lines.eps',
    out_png: str = 'lines.png'
    
) -> None:
    """Compute the line segments using DeepLSD. Adaptation of the original author's notebook.

    Args:
        im: Path to the input image.
        ckpt: Path to the model weights.
        out_txt: Path for the output TXT.
        out_svg: Path for the output SVG.
        out_png: Path for the output PNG.
    """
    # Load an image
    img = cv2.imread(im)[:, :, ::-1]
    gray_img = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)
    h, w, c = img.shape
    
    # Model config
    device = torch.device('cpu')
    conf = {
        'sharpen': True,  # Use the DF normalization (should be True)
        'detect_lines': True,  # Whether to detect lines or only DF/AF
        'line_detection_params': {
            'merge': False,  # Whether to merge close-by lines
            'optimize': False,  # Whether to refine the lines after detecting them
            'use_vps': True,  # Whether to use vanishing points (VPs) in the refinement
            'optimize_vps': True,  # Whether to also refine the VPs in the refinement
            'filtering': True,  # Whether to filter out lines based on the DF/AF. Use 'strict' to get an even stricter filtering
            'grad_thresh': 3,
            'grad_nfa': True,  # If True, use the image gradient and the NFA score of LSD to further threshold lines. We recommand using it for easy images, but to turn it off for challenging images (e.g. night, foggy, blurry images)
        }
    }
    
    # Load the model
    ckpt = torch.load(str(ckpt), map_location='cpu')
    net = DeepLSD(conf)
    net.load_state_dict(ckpt['model'])
    net = net.to(device).eval()
    
    # Detect (and optionally refine) the lines
    inputs = {'image': torch.tensor(gray_img, dtype=torch.float, device=device)[None, None] / 255.}
    with torch.no_grad():
        out = net(inputs)
        pred_lines = out['lines'][0]
    
    ##### OUTPUT ######
    
    ftxt = open(out_txt, "w")
    fcsv = open(out_svg, "w")
    feps = open(out_eps, "w")

    fcsv.write('<?xml version="1.0" standalone="no"?>\n<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN"\n"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd"><svg version="1.1" xmlns="http://www.w3.org/2000/svg"\nxmlns:xlink="http://www.w3.org/1999/xlink"\nwidth="'+str(w)+'" height="'+str(h)+'" viewBox="0 0 '+str(w)+' '+str(h)+'">')

    feps.write("%!PS-Adobe-3.0 EPSF-3.0\n");
    feps.write(f"%%BoundingBox: 0 0 {w} {h}\n");
    feps.write("%%Creator: LSD, Line Segment Detector\n");
    feps.write(f"%%Title: ({out_eps})\n");
    feps.write("%%EndComments\n");

    color='black'
    width=1.5
    for tp_id, line in enumerate(pred_lines):
        x1, y1 = line[0] # this is yx
        x2, y2 = line[1] # this is yx
        fcsv.write('<line x1="'+str(x1.item())+'" y1="' +str(y1.item())+ '" x2="' +str(x2.item())+ '" y2="' +str(y2.item())+ '" stroke="'+color+'" stroke-width="'+str(width)+'"/>\n')
        ftxt.write(str(x1.item())+' '+str(y1.item())+' '+str(x2.item())+' '+str(y2.item())+'\n')
        feps.write(f"newpath {x1} {h-y1} moveto {x2} {h-y2} lineto 1.5 setlinewidth stroke\n")

    fcsv.write('\n</svg>')
    feps.write("showpage\n");
    feps.write("%%EOF\n");

    ftxt.close()
    fcsv.close()
    feps.close()
    
    p2 = ['gs', '-dNOPAUSE', '-dBATCH', '-sDEVICE=pnggray', '-dGraphicsAlphaBits=4', '-r72', \
            '-dEPSCrop', '-sOutputFile='+out_png, out_eps]
    val1 = subprocess.run(p2)
    im = iio.read(out_png)
    im = im.mean(axis=-1)
    im = 255-im
    iio.write(out_png, im)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument('--im', type=str, help='image path')
    parser.add_argument('--ckpt', type=str, help='path to checkpoint')

    opts = parser.parse_args()
    main(opts.im, opts.ckpt)
